@@ -30,6 +30,7 @@
 struct nl_cache_ops;
 struct nl_sock;
 struct nl_object;
+struct nl_hash_table;
 
 struct nl_cb
 {
@@ -78,6 +79,9 @@ struct nl_cache
 	int			c_nitems;
 	int                     c_iarg1;
 	int                     c_iarg2;
+	int			c_refcnt;
+	unsigned int		c_flags;
+	struct nl_hash_table *	hashtable;
 	struct nl_cache_ops *   c_ops;
 };
 
@@ -94,6 +98,7 @@ struct nl_cache_mngr
 	int			cm_flags;
 	int			cm_nassocs;
 	struct nl_sock *	cm_sock;
+	struct nl_sock *	cm_sync_sock;
 	struct nl_cache_assoc *	cm_assocs;
 };
 
@@ -177,6 +182,7 @@ struct rtnl_link
 	uint32_t			l_num_tx_queues;
 	uint32_t			l_num_rx_queues;
 	uint32_t			l_group;
+	uint8_t				l_carrier;
 };
 
 struct rtnl_ncacheinfo
@@ -202,6 +208,7 @@ struct rtnl_neigh
 	struct rtnl_ncacheinfo n_cacheinfo;
 	uint32_t                n_state_mask;
 	uint32_t                n_flag_mask;
+	uint32_t		n_master;
 };
 
 
@@ -784,6 +791,50 @@ struct nfnl_ct {
 	struct nfnl_ct_dir	ct_repl;
 };
 
+union nfnl_exp_protodata {
+	struct {
+		uint16_t	src;
+		uint16_t	dst;
+	} port;
+	struct {
+		uint16_t	id;
+		uint8_t		type;
+		uint8_t		code;
+	} icmp;
+};
+
+// Allow for different master/expect l4 protocols
+struct nfnl_exp_proto
+{
+	uint8_t						l4protonum;
+	union nfnl_exp_protodata	l4protodata;
+};
+
+struct nfnl_exp_dir {
+	struct nl_addr *		src;
+	struct nl_addr *		dst;
+	struct nfnl_exp_proto	proto;
+};
+
+struct nfnl_exp {
+	NLHDR_COMMON
+
+	uint8_t			exp_family;
+	uint32_t		exp_timeout;
+	uint32_t		exp_id;
+	uint16_t		exp_zone;
+	uint32_t		exp_class;
+	uint32_t		exp_flags;
+	char *			exp_helper_name;
+	char *			exp_fn;
+	uint8_t			exp_nat_dir;
+
+	struct nfnl_exp_dir		exp_expect;
+	struct nfnl_exp_dir		exp_master;
+	struct nfnl_exp_dir		exp_mask;
+	struct nfnl_exp_dir		exp_nat;
+};
+
 struct nfnl_log {
 	NLHDR_COMMON
 
