@@ -6,156 +6,24 @@
  *	License as published by the Free Software Foundation version 2.1
  *	of the License.
  *
- * Copyright (c) 2003-2006 Thomas Graf <tgraf@suug.ch>
+ * Copyright (c) 2003-2012 Thomas Graf <tgraf@suug.ch>
  */
 
 /**
- * @ingroup nl
- * @defgroup msg Messages
+ * @ingroup core
+ * @defgroup msg Message Construction & Parsing
  * Netlink Message Construction/Parsing Interface
  * 
- * The following information is partly extracted from RFC3549
- * (ftp://ftp.rfc-editor.org/in-notes/rfc3549.txt)
+ * Related sections in the development guide:
+ * - @core_doc{_message_parsing_amp_construction,Message Parsing & Construction}
  *
- * @par Message Format
- * Netlink messages consist of a byte stream with one or multiple
- * Netlink headers and an associated payload.  If the payload is too big
- * to fit into a single message it, can be split over multiple Netlink
- * messages, collectively called a multipart message.  For multipart
- * messages, the first and all following headers have the \c NLM_F_MULTI
- * Netlink header flag set, except for the last header which has the
- * Netlink header type \c NLMSG_DONE.
- *
- * @par
- * The Netlink message header (\link nlmsghdr struct nlmsghdr\endlink) is shown below.
- * @code   
- * 0                   1                   2                   3
- * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * |                          Length                             |
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * |            Type              |           Flags              |
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * |                      Sequence Number                        |
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * |                      Process ID (PID)                       |
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * @endcode
- *
- * @par
- * The netlink message header and payload must be aligned properly:
- * @code
- *  <------- NLMSG_ALIGN(hlen) ------> <---- NLMSG_ALIGN(len) --->
- * +----------------------------+- - -+- - - - - - - - - - -+- - -+
- * |           Header           | Pad |       Payload       | Pad |
- * |      struct nlmsghdr       |     |                     |     |
- * +----------------------------+- - -+- - - - - - - - - - -+- - -+
- * @endcode
- * @par
- * Message Format:
- * @code
- *    <--- nlmsg_total_size(payload)  --->
- *    <-- nlmsg_msg_size(payload) ->
- *   +----------+- - -+-------------+- - -+-------- - -
- *   | nlmsghdr | Pad |   Payload   | Pad | nlmsghdr
- *   +----------+- - -+-------------+- - -+-------- - -
- *   nlmsg_data(nlh)---^                   ^
- *   nlmsg_next(nlh)-----------------------+
- * @endcode
- * @par
- * The payload may consist of arbitary data but may have strict
- * alignment and formatting rules depening on the specific netlink
- * families.
- * @par
- * @code
- *    <---------------------- nlmsg_len(nlh) --------------------->
- *    <------ hdrlen ------>       <- nlmsg_attrlen(nlh, hdrlen) ->
- *   +----------------------+- - -+--------------------------------+
- *   |     Family Header    | Pad |           Attributes           |
- *   +----------------------+- - -+--------------------------------+
- *   nlmsg_attrdata(nlh, hdrlen)---^
- * @endcode
- * @par The ACK Netlink Message
- * This message is actually used to denote both an ACK and a NACK.
- * Typically, the direction is from FEC to CPC (in response to an ACK
- * request message).  However, the CPC should be able to send ACKs back
- * to FEC when requested.
- * @code
- *  0                   1                   2                   3
- *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * |                       Netlink message header                  |
- * |                       type = NLMSG_ERROR                      |
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * |                          Error code                           |
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * |                       OLD Netlink message header              |
- * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
- * @endcode
- *
- * @par Example
- * @code
- * // Various methods exist to create/allocate a new netlink
- * // message. 
- * //
- * // nlmsg_alloc() will allocate an empty netlink message with
- * // a maximum payload size which defaults to the page size of
- * // the system. This default size can be modified using the
- * // function nlmsg_set_default_size().
- * struct nl_msg *msg = nlmsg_alloc();
- *
- * // Very often, the message type and message flags are known
- * // at allocation time while the other fields are auto generated:
- * struct nl_msg *msg = nlmsg_alloc_simple(MY_TYPE, MY_FLAGS);
- *
- * // Alternatively an existing netlink message header can be used
- * // to inherit the header values:
- * struct nlmsghdr hdr = {
- * 	.nlmsg_type = MY_TYPE,
- * 	.nlmsg_flags = MY_FLAGS,
- * };
- * struct nl_msg *msg = nlmsg_inherit(&hdr);
- *
- * // Last but not least, netlink messages received from netlink sockets
- * // can be converted into nl_msg objects using nlmsg_convert(). This
- * // will create a message with a maximum payload size which equals the
- * // length of the existing netlink message, therefore no more data can
- * // be appened without calling nlmsg_expand() first.
- * struct nl_msg *msg = nlmsg_convert(nlh_from_nl_sock);
- *
- * // Payload may be added to the message via nlmsg_append(). The fourth
- * // parameter specifies the number of alignment bytes the data should
- * // be padding with at the end. Common values are 0 to disable it or
- * // NLMSG_ALIGNTO to ensure proper netlink message padding.
- * nlmsg_append(msg, &mydata, sizeof(mydata), 0);
- *
- * // Sometimes it may be necessary to reserve room for data but defer
- * // the actual copying to a later point, nlmsg_reserve() can be used
- * // for this purpose:
- * void *data = nlmsg_reserve(msg, sizeof(mydata), NLMSG_ALIGNTO);
- *
- * // Attributes may be added using the attributes interface.
- *
- * // After successful use of the message, the memory must be freed
- * // using nlmsg_free()
- * nlmsg_free(msg);
- * @endcode
- * 
- * @par 4) Parsing messages
- * @code
- * int n;
- * unsigned char *buf;
- * struct nlmsghdr *hdr;
- *
- * n = nl_recv(handle, NULL, &buf);
- * 
- * hdr = (struct nlmsghdr *) buf;
- * while (nlmsg_ok(hdr, n)) {
- * 	// Process message here...
- * 	hdr = nlmsg_next(hdr, &n);
- * }
- * @endcode
  * @{
+ *
+ * Header
+ * ------
+ * ~~~~{.c}
+ * #include <netlink/msg.h>
+ * ~~~~
  */
 
 #include <netlink-local.h>
@@ -178,17 +46,28 @@ static void __init init_msg_size(void)
  */
 
 /**
- * length of netlink message not including padding
- * @arg payload		length of message payload
+ * Calculates size of netlink message based on payload length.
+ * @arg payload		Length of payload
+ *
+ * @return size of netlink message without padding.
  */
-int nlmsg_msg_size(int payload)
+int nlmsg_size(int payload)
 {
 	return NLMSG_HDRLEN + payload;
 }
 
+static int nlmsg_msg_size(int payload)
+{
+	return nlmsg_size(payload);
+}
+
 /**
- * length of netlink message including padding
- * @arg payload		length of message payload
+ * Calculates size of netlink message including padding based on payload length
+ * @arg payload		Length of payload
+ *
+ * This function is idential to nlmsg_size() + nlmsg_padlen().
+ *
+ * @return Size of netlink message including padding.
  */
 int nlmsg_total_size(int payload)
 {
@@ -196,8 +75,14 @@ int nlmsg_total_size(int payload)
 }
 
 /**
- * length of padding at the message's tail
- * @arg payload		length of message payload
+ * Size of padding that needs to be added at end of message
+ * @arg payload		Length of payload
+ *
+ * Calculates the number of bytes of padding which is required to be added to
+ * the end of the message to ensure that the next netlink message header begins
+ * properly aligned to NLMSG_ALIGNTO.
+ *
+ * @return Number of bytes of padding needed.
  */
 int nlmsg_padlen(int payload)
 {
@@ -207,13 +92,15 @@ int nlmsg_padlen(int payload)
 /** @} */
 
 /**
- * @name Payload Access
+ * @name Access to Message Payload
  * @{
  */
 
 /**
- * head of message payload
- * @arg nlh		netlink messsage header
+ * Return pointer to message payload
+ * @arg nlh		Netlink message header
+ *
+ * @return Pointer to start of message payload.
  */
 void *nlmsg_data(const struct nlmsghdr *nlh)
 {
@@ -226,14 +113,21 @@ void *nlmsg_tail(const struct nlmsghdr *nlh)
 }
 
 /**
- * length of message payload
- * @arg nlh		netlink message header
+ * Return length of message payload
+ * @arg nlh		Netlink message header
+ *
+ * @return Length of message payload in bytes.
  */
-int nlmsg_len(const struct nlmsghdr *nlh)
+int nlmsg_datalen(const struct nlmsghdr *nlh)
 {
 	return nlh->nlmsg_len - NLMSG_HDRLEN;
 }
 
+static int nlmsg_len(const struct nlmsghdr *nlh)
+{
+	return nlmsg_datalen(nlh);
+}
+
 /** @} */
 
 /**
@@ -284,7 +178,7 @@ int nlmsg_valid_hdr(const struct nlmsghdr *nlh, int hdrlen)
  */
 int nlmsg_ok(const struct nlmsghdr *nlh, int remaining)
 {
-	return (remaining >= sizeof(struct nlmsghdr) &&
+	return (remaining >= (int)sizeof(struct nlmsghdr) &&
 		nlh->nlmsg_len >= sizeof(struct nlmsghdr) &&
 		nlh->nlmsg_len <= remaining);
 }
@@ -320,7 +214,7 @@ int nlmsg_parse(struct nlmsghdr *nlh, int hdrlen, struct nlattr *tb[],
 		int maxtype, struct nla_policy *policy)
 {
 	if (!nlmsg_valid_hdr(nlh, hdrlen))
-		return nl_errno(EINVAL);
+		return -NLE_MSG_TOOSHORT;
 
 	return nla_parse(tb, maxtype, nlmsg_attrdata(nlh, hdrlen),
 			 nlmsg_attrlen(nlh, hdrlen), policy);
@@ -351,7 +245,7 @@ int nlmsg_validate(struct nlmsghdr *nlh, int hdrlen, int maxtype,
 		   struct nla_policy *policy)
 {
 	if (!nlmsg_valid_hdr(nlh, hdrlen))
-		return nl_errno(EINVAL);
+		return -NLE_MSG_TOOSHORT;
 
 	return nla_validate(nlmsg_attrdata(nlh, hdrlen),
 			    nlmsg_attrlen(nlh, hdrlen), maxtype, policy);
@@ -368,14 +262,21 @@ static struct nl_msg *__nlmsg_alloc(size_t len)
 {
 	struct nl_msg *nm;
 
+	if (len < sizeof(struct nlmsghdr))
+		len = sizeof(struct nlmsghdr);
+
 	nm = calloc(1, sizeof(*nm));
 	if (!nm)
 		goto errout;
 
+	nm->nm_refcnt = 1;
+
 	nm->nm_nlh = calloc(1, len);
 	if (!nm->nm_nlh)
 		goto errout;
 
+	memset(nm->nm_nlh, 0, sizeof(struct nlmsghdr));
+
 	nm->nm_protocol = -1;
 	nm->nm_size = len;
 	nm->nm_nlh->nlmsg_len = nlmsg_total_size(0);
@@ -385,7 +286,6 @@ static struct nl_msg *__nlmsg_alloc(size_t len)
 	return nm;
 errout:
 	free(nm);
-	nl_errno(ENOMEM);
 	return NULL;
 }
 
@@ -517,10 +417,8 @@ void *nlmsg_reserve(struct nl_msg *n, size_t len, int pad)
 
 	tlen = pad ? ((len + (pad - 1)) & ~(pad - 1)) : len;
 
-	if ((tlen + nlmsg_len) > n->nm_size) {
-		nl_errno(ENOBUFS);
+	if ((tlen + nlmsg_len) > n->nm_size)
 		return NULL;
-	}
 
 	buf += nlmsg_len;
 	n->nm_nlh->nlmsg_len += tlen;
@@ -528,8 +426,8 @@ void *nlmsg_reserve(struct nl_msg *n, size_t len, int pad)
 	if (tlen > len)
 		memset(buf + len, 0, tlen - len);
 
-	NL_DBG(2, "msg %p: Reserved %zu bytes, pad=%d, nlmsg_len=%d\n",
-		  n, len, pad, n->nm_nlh->nlmsg_len);
+	NL_DBG(2, "msg %p: Reserved %zu (%zu) bytes, pad=%d, nlmsg_len=%d\n",
+		  n, tlen, len, pad, n->nm_nlh->nlmsg_len);
 
 	return buf;
 }
@@ -552,7 +450,7 @@ int nlmsg_append(struct nl_msg *n, void *data, size_t len, int pad)
 
 	tmp = nlmsg_reserve(n, len, pad);
 	if (tmp == NULL)
-		return nl_errno(ENOMEM);
+		return -NLE_NOMEM;
 
 	memcpy(tmp, data, len);
 	NL_DBG(2, "msg %p: Appended %zu bytes with padding %d\n", n, len, pad);
@@ -579,11 +477,11 @@ int nlmsg_expand(struct nl_msg *n, size_t newlen)
 	void *tmp;
 
 	if (newlen <= n->nm_size)
-		return nl_errno(EINVAL);
+		return -NLE_INVAL;
 
 	tmp = realloc(n->nm_nlh, newlen);
 	if (tmp == NULL)
-		return nl_errno(ENOMEM);
+		return -NLE_NOMEM;
 
 	n->nm_nlh = tmp;
 	n->nm_size = newlen;
@@ -646,21 +544,39 @@ struct nlmsghdr *nlmsg_hdr(struct nl_msg *n)
 }
 
 /**
- * Free a netlink message
- * @arg n		netlink message
- *
- * Destroys a netlink message and frees up all used memory.
- *
- * @pre The message must be unused.
+ * Acquire a reference on a netlink message
+ * @arg msg		message to acquire reference from
  */
-void nlmsg_free(struct nl_msg *n)
+void nlmsg_get(struct nl_msg *msg)
 {
-	if (!n)
+	msg->nm_refcnt++;
+	NL_DBG(4, "New reference to message %p, total %d\n",
+	       msg, msg->nm_refcnt);
+}
+
+/**
+ * Release a reference from an netlink message
+ * @arg msg		message to release reference from
+ *
+ * Frees memory after the last reference has been released.
+ */
+void nlmsg_free(struct nl_msg *msg)
+{
+	if (!msg)
 		return;
 
-	free(n->nm_nlh);
-	free(n);
-	NL_DBG(2, "msg %p: Freed\n", n);
+	msg->nm_refcnt--;
+	NL_DBG(4, "Returned message reference %p, %d remaining\n",
+	       msg, msg->nm_refcnt);
+
+	if (msg->nm_refcnt < 0)
+		BUG();
+
+	if (msg->nm_refcnt <= 0) {
+		free(msg->nm_nlh);
+		free(msg);
+		NL_DBG(2, "msg %p: Freed\n", msg);
+	}
 }
 
 /** @} */
@@ -725,7 +641,7 @@ struct ucred *nlmsg_get_creds(struct nl_msg *msg)
  * @{
  */
 
-static struct trans_tbl nl_msgtypes[] = {
+static const struct trans_tbl nl_msgtypes[] = {
 	__ADD(NLMSG_NOOP,NOOP)
 	__ADD(NLMSG_ERROR,ERROR)
 	__ADD(NLMSG_DONE,DONE)
@@ -817,15 +733,18 @@ int nl_msg_parse(struct nl_msg *msg, void (*cb)(struct nl_object *, void *),
 		.cb = cb,
 		.arg = arg,
 	};
+	int err;
 
-	ops = nl_cache_ops_associate(nlmsg_get_proto(msg),
-				     nlmsg_hdr(msg)->nlmsg_type);
+	ops = nl_cache_ops_associate_safe(nlmsg_get_proto(msg),
+					  nlmsg_hdr(msg)->nlmsg_type);
 	if (ops == NULL)
-		return nl_error(ENOENT, "Unknown message type %d",
-				nlmsg_hdr(msg)->nlmsg_type);
+		return -NLE_MSGTYPE_NOSUPPORT;
 	p.pp_arg = &x;
 
-	return nl_cache_parse(ops, NULL, nlmsg_hdr(msg), &p);
+	err = nl_cache_parse(ops, NULL, nlmsg_hdr(msg), &p);
+	nl_cache_ops_put(ops);
+
+	return err;
 }
 
 /** @} */
@@ -886,13 +805,14 @@ static void print_hdr(FILE *ofd, struct nl_msg *msg)
 
 	fprintf(ofd, "    .nlmsg_len = %d\n", nlh->nlmsg_len);
 
-	ops = nl_cache_ops_associate(nlmsg_get_proto(msg), nlh->nlmsg_type);
+	ops = nl_cache_ops_associate_safe(nlmsg_get_proto(msg), nlh->nlmsg_type);
 	if (ops) {
 		mt = nl_msgtype_lookup(ops, nlh->nlmsg_type);
 		if (!mt)
 			BUG();
 
 		snprintf(buf, sizeof(buf), "%s::%s", ops->co_name, mt->mt_name);
+		nl_cache_ops_put(ops);
 	} else
 		nl_nlmsgtype2str(nlh->nlmsg_type, buf, sizeof(buf));
 
@@ -952,7 +872,7 @@ void nl_msg_dump(struct nl_msg *msg, FILE *ofd)
 	"--------------------------   BEGIN NETLINK MESSAGE "
 	"---------------------------\n");
 
-	fprintf(ofd, "  [HEADER] %Zu octets\n", sizeof(struct nlmsghdr));
+	fprintf(ofd, "  [HEADER] %zu octets\n", sizeof(struct nlmsghdr));
 	print_hdr(ofd, msg);
 
 	if (hdr->nlmsg_type == NLMSG_ERROR &&
@@ -960,10 +880,10 @@ void nl_msg_dump(struct nl_msg *msg, FILE *ofd)
 		struct nl_msg *errmsg;
 		struct nlmsgerr *err = nlmsg_data(hdr);
 
-		fprintf(ofd, "  [ERRORMSG] %Zu octets\n", sizeof(*err));
+		fprintf(ofd, "  [ERRORMSG] %zu octets\n", sizeof(*err));
 		fprintf(ofd, "    .error = %d \"%s\"\n", err->error,
 			strerror(-err->error));
-		fprintf(ofd, "  [ORIGINAL MESSAGE] %Zu octets\n", sizeof(*hdr));
+		fprintf(ofd, "  [ORIGINAL MESSAGE] %zu octets\n", sizeof(*hdr));
 
 		errmsg = nlmsg_inherit(&err->msg);
 		print_hdr(ofd, errmsg);
@@ -973,8 +893,8 @@ void nl_msg_dump(struct nl_msg *msg, FILE *ofd)
 		int payloadlen = nlmsg_len(hdr);
 		int attrlen = 0;
 
-		ops = nl_cache_ops_associate(nlmsg_get_proto(msg),
-					     hdr->nlmsg_type);
+		ops = nl_cache_ops_associate_safe(nlmsg_get_proto(msg),
+						  hdr->nlmsg_type);
 		if (ops) {
 			attrlen = nlmsg_attrlen(hdr, ops->co_hdrsize);
 			payloadlen -= attrlen;
@@ -991,6 +911,9 @@ void nl_msg_dump(struct nl_msg *msg, FILE *ofd)
 			attrlen = nlmsg_attrlen(hdr, ops->co_hdrsize);
 			dump_attrs(ofd, attrs, attrlen, 0);
 		}
+
+		if (ops)
+			nl_cache_ops_put(ops);
 	}
 
 	fprintf(ofd, 
